# FastApi sqlalchemy exemple avancé

![tag](fastapi)
![category](python)

cet exemple fournit 4 apis

on peut voir que l'usage de PersonFlatOut permet d'avoir une sortie du type

```json
[
  {
    "id": 0,
    "name": "string",
    "sex": {
      "id": 0,
      "lib": "string"
    }
  }
]
```

au lieu de 


```json
[
  {
	"id": 1,
	"name": "Me",
	"sex": "Man"
  }
]
```


```python
from fastapi import FastAPI, HTTPException, Depends
from pydantic import BaseModel, ConfigDict
from typing import List

from sqlalchemy import create_engine, Column, Integer, String, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker, Session, joinedload

# --- SQLAlchemy setup ---
DATABASE_URL = "sqlite:///./test.db"
engine = create_engine(DATABASE_URL, echo=True)
Base = declarative_base()
SessionLocal = sessionmaker(bind=engine)


class Sex(Base):
    __tablename__ = 'sex'
    id = Column(Integer, primary_key=True)
    lib = Column(String)
    persons = relationship("Person", back_populates="sex")


class Person(Base):
    __tablename__ = 'person'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    sex_id = Column(Integer, ForeignKey('sex.id'))
    
    sex = relationship("Sex", back_populates="persons")


Base.metadata.create_all(engine)

def init_db():
    db = SessionLocal()
    if db.query(Sex).count() == 0:
        s1 = Sex(lib="Man")
        s2 = Sex(lib="Woman")
        db.add_all([s1, s2])
        db.commit()

        p1 = Person(name="Me", sex=s1)
        p2 = Person(name="Her", sex=s2)
        db.add_all([p1, p2])
        db.commit()
    db.close()

# --- Pydantic schemas ---
class SexOut(BaseModel):
    id: int
    lib: str

    model_config = ConfigDict(from_attributes=True)

class PersonOut(BaseModel):
    id: int
    name: str
    sex: SexOut
    
    model_config = ConfigDict(from_attributes=True)

class PersonFlatOut(BaseModel):
    id: int
    name: str
    sex: str

    model_config = ConfigDict(from_attributes=True)

    @classmethod
    def from_orm_obj(cls, person: Person) -> "PersonFlatOut":
        return cls(
            id=person.id,
            name=person.name,
            sex=person.sex.lib if person.sex else None
        )

# --- FastAPI app ---
app = FastAPI()

# Dependency to get DB session
def get_db():
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()

@app.get("/sexs", response_model=List[SexOut])
def get_sexs(db: Session = Depends(get_db)):
    return db.query(Sex).all()

@app.get("/sex/{sex_id}", response_model=SexOut)
def get_sex(sex_id: int, db: Session = Depends(get_db)):
    sex = db.query(Sex).filter(Sex.id == sex_id).first()
    if not sex:
        raise HTTPException(status_code=404, detail="No Sex found for this sex ID")
    return sex

@app.get("/persons", response_model=List[PersonFlatOut])
def get_persons(db: Session = Depends(get_db)):
    #return db.query(Person).all()
    persons = db.query(Person).all()
    return [PersonFlatOut.from_orm_obj(p) for p in persons]

@app.get("/person/{person_id}", response_model=PersonOut)
def get_person(person_id: int, db: Session = Depends(get_db)):
    person = db.query(Person).filter(Person.id == person_id).first()
    if not person:
        raise HTTPException(status_code=404, detail="No Person found for this person ID")
    return person

@app.get("/personflat/{person_id}", response_model=PersonFlatOut)
def get_personflat(person_id: int, db: Session = Depends(get_db)):
    person = db.query(Person).filter(Person.id == person_id).first()
    if not person:
        raise HTTPException(status_code=404, detail="No Person found for this person ID")
    
    return PersonFlatOut.from_orm_obj(person)



if __name__ == "__main__":
    init_db()
    import uvicorn
    uvicorn.run("testapi:app", port=8080, reload=True)
```

Une autre solution que je préfère est d'utiliser des computed_field.

Cette solution a plusieurs avantages:

- au niveau de l'api on génère toujours des objets de la base (Pas besoin de générer des elements pydantics avec un from_orm_obj
- la gestion des champs est géré au niveau pydantic et centralisé
- on peut exclure des objets au moment de la réponse


```python
from fastapi import FastAPI, HTTPException, Depends
from pydantic import BaseModel, ConfigDict, computed_field, Field
from typing import List
from typing import Optional
from typing_extensions import Annotated

from sqlalchemy import create_engine, Column, Integer, String, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker, Session, joinedload

# --- SQLAlchemy setup ---
DATABASE_URL = "sqlite:///./test.db"
engine = create_engine(DATABASE_URL, echo=True)
Base = declarative_base()
SessionLocal = sessionmaker(bind=engine)


class Sex(Base):
    __tablename__ = 'sex'
    id = Column(Integer, primary_key=True)
    lib = Column(String)
    persons = relationship("Person", back_populates="sex")


class Person(Base):
    __tablename__ = 'person'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    sex_id = Column(Integer, ForeignKey('sex.id'))    
    sex = relationship("Sex", back_populates="persons")


Base.metadata.create_all(engine)

def init_db():
    db = SessionLocal()
    if db.query(Sex).count() == 0:
        s1 = Sex(lib="Man")
        s2 = Sex(lib="Woman")
        db.add_all([s1, s2])
        db.commit()

        p1 = Person(name="Me", sex=s1)
        p2 = Person(name="Her", sex=s2)
        db.add_all([p1, p2])
        db.commit()
    db.close()

# --- Pydantic schemas ---
class SexOut(BaseModel):
    id: int
    lib: str

    model_config = ConfigDict(from_attributes=True)

class PersonOut(BaseModel):
    id: int
    name: str
    sex: SexOut
    
    model_config = ConfigDict(from_attributes=True)

class PersonFlatOut(BaseModel):
    id: int
    name: str
    sex: SexOut = Field(exclude=True)

    model_config = ConfigDict(from_attributes=True)
    
    @computed_field
    def sexLib(self) -> Annotated[str, str]:
        return self.sex.lib if self.sex else ""

# --- FastAPI app ---
app = FastAPI()

# Dependency to get DB session
def get_db():
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()

@app.get("/sexs", response_model=List[SexOut])
def get_sexs(db: Session = Depends(get_db)):
    return db.query(Sex).all()

@app.get("/sex/{sex_id}", response_model=SexOut)
def get_sex(sex_id: int, db: Session = Depends(get_db)):
    sex = db.query(Sex).filter(Sex.id == sex_id).first()
    if not sex:
        raise HTTPException(status_code=404, detail="No Sex found for this sex ID")
    return sex

@app.get("/persons", response_model=List[PersonFlatOut])
def get_persons(db: Session = Depends(get_db)):
    return db.query(Person).all()

@app.get("/person/{person_id}", response_model=PersonOut)
def get_person(person_id: int, db: Session = Depends(get_db)):
    person = db.query(Person).filter(Person.id == person_id).first()
    if not person:
        raise HTTPException(status_code=404, detail="No Person found for this person ID")
    return person

@app.get("/personflat/{person_id}", response_model=PersonFlatOut)
def get_personflat(person_id: int, db: Session = Depends(get_db)):
    person = db.query(Person).options(joinedload(Person.sex)).filter(Person.id == person_id).first()
    if not person:
        raise HTTPException(status_code=404, detail="No Person found for this person ID")
    return person



if __name__ == "__main__":
    init_db()
    import uvicorn
    uvicorn.run("testapi:app", port=8080, reload=True)
```

une dernière modification qui permet de simplifier l'accès aux champs calculés (qui les génèrent dynamiquement)

```python
from fastapi import FastAPI, HTTPException, Depends
from pydantic import BaseModel, ConfigDict, computed_field, Field, ConfigDict
from typing import get_type_hints, Annotated, get_args, get_origin, List, Optional
from types import FunctionType

from sqlalchemy import create_engine, Column, Integer, String, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker, Session, joinedload

# --- SQLAlchemy setup ---
DATABASE_URL = "sqlite:///./test.db"
engine = create_engine(DATABASE_URL, echo=True)
Base = declarative_base()
SessionLocal = sessionmaker(bind=engine)


class Sex(Base):
    __tablename__ = 'sex'
    id = Column(Integer, primary_key=True)
    lib = Column(String)
    persons = relationship("Person", back_populates="sex")


class Person(Base):
    __tablename__ = 'person'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    sex_id = Column(Integer, ForeignKey('sex.id'))    
    sex = relationship("Sex", back_populates="persons")


Base.metadata.create_all(engine)

def init_db():
    db = SessionLocal()
    if db.query(Sex).count() == 0:
        s1 = Sex(lib="Man")
        s2 = Sex(lib="Woman")
        db.add_all([s1, s2])
        db.commit()

        p1 = Person(name="Me", sex=s1)
        p2 = Person(name="Her", sex=s2)
        db.add_all([p1, p2])
        db.commit()
    db.close()

# --- Pydantic schemas ---
class DynamicComputed:
    def __init__(self, attr_path: str):
        self.attr_path = attr_path

class DynamicComputedModel(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    def __init_subclass__(cls):
        super().__init_subclass__()

        computed = getattr(cls, "__computed_fields__", {})

        for field_name, config in computed.items():
            if hasattr(cls, field_name):
                raise TypeError(
                    f"Field '{field_name}' already exists in {cls.__name__}. "
                    "Cannot define a field and generate a computed_field with the same name."
                )

            # Supporte les 2 formats : string simple ou (path, type)
            if isinstance(config, tuple):
                attr_path, return_type = config
            else:
                attr_path = config
                return_type = str  # défaut

            def make_func(path):
                def f(self):
                    val = self
                    for part in path.split("."):
                        val = getattr(val, part, None)
                        if val is None:
                            return None
                    return val
                return f

            func = make_func(attr_path)
            func.__name__ = field_name
            setattr(cls, field_name, computed_field(func, return_type=return_type))

class SexOut(BaseModel):
    id: int
    lib: str

    model_config = ConfigDict(from_attributes=True)

class PersonOut(BaseModel):
    id: int
    name: str
    sex: SexOut
    
    model_config = ConfigDict(from_attributes=True)

class PersonFlatOut(DynamicComputedModel):
    id: int
    name: str
    sex: SexOut = Field(exclude=True)

    __computed_fields__ = {
        "sexId": ("sex.id", int),
        "sexLib": ("sex.lib", str),
    }

# --- FastAPI app ---
app = FastAPI()

# Dependency to get DB session
def get_db():
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()

@app.get("/sexs", response_model=List[SexOut])
def get_sexs(db: Session = Depends(get_db)):
    return db.query(Sex).all()

@app.get("/sex/{sex_id}", response_model=SexOut)
def get_sex(sex_id: int, db: Session = Depends(get_db)):
    sex = db.query(Sex).filter(Sex.id == sex_id).first()
    if not sex:
        raise HTTPException(status_code=404, detail="No Sex found for this sex ID")
    return sex

@app.get("/persons", response_model=List[PersonFlatOut])
def get_persons(db: Session = Depends(get_db)):
    return db.query(Person).all()

@app.get("/person/{person_id}", response_model=PersonOut)
def get_person(person_id: int, db: Session = Depends(get_db)):
    person = db.query(Person).filter(Person.id == person_id).first()
    if not person:
        raise HTTPException(status_code=404, detail="No Person found for this person ID")
    return person

@app.get("/personflat/{person_id}", response_model=PersonFlatOut)
def get_personflat(person_id: int, db: Session = Depends(get_db)):
    person = db.query(Person).options(joinedload(Person.sex)).filter(Person.id == person_id).first()
    if not person:
        raise HTTPException(status_code=404, detail="No Person found for this person ID")
    return person



if __name__ == "__main__":
    init_db()
    import uvicorn
    uvicorn.run("testapi:app", port=8080, reload=True)
```


